/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls/cpp/filesearch.hpp
 *
 * Purpose:     recls C++ mapping - FileSearch class.
 *
 * Created:     18th August 2003
 * Updated:     15th March 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_RECLS_CPP_HPP_FILESEARCH
#define RECLS_INCL_RECLS_CPP_HPP_FILESEARCH

/* File version */
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_RECLS_CPP_HPP_FILESEARCH_MAJOR       2
# define RECLS_VER_RECLS_CPP_HPP_FILESEARCH_MINOR       2
# define RECLS_VER_RECLS_CPP_HPP_FILESEARCH_REVISION    1
# define RECLS_VER_RECLS_CPP_HPP_FILESEARCH_EDIT        17
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file recls/cpp/filesearch.hpp Definition of the \link recls::cpp::FileSearch FileSearch\endlink class, part of 
 * the \ref group_recls_cpp "recls C++ mapping"
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/cpp/recls.hpp>
#include <recls/cpp/search.hpp>
#include <recls_assert.h>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{

namespace cpp
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// This class provides functionality for executing a file search
///
/// \ingroup group_recls_cpp
class FileSearch
    : public Search
{
/// \name Types
public:
/// @{
    typedef recls_char_t    char_type;
    typedef FileSearch      class_type;
/// @}

/// \name Construction
/// @{
public:
    /// Creates a search for the items identified by \c pattern in the directory \c rootDir, according to the given \c flags.
    ///
    /// \param rootDir The starting directory for the search. If NULL, or the empty string, uses the current directory
    /// \param pattern The search pattern, e.g. "*.h". If NULL, or the empty string, uses "*.*" (Win32) or "*" (UNIX)
    /// \param flags A combination of enumerants from the \c RECLS_FLAG enumeration
    FileSearch(char_type const *rootDir, char_type const *pattern, recls_uint32_t flags);
    /// Destructor
    ~FileSearch();
/// @}

/// \name Operations
/// @{
public:
#if defined(RECLS_DOCUMENTATION_SKIP_SECTION) || \
    (   RECLS_VER_MAJOR > 1 || \
        RECLS_VER_MINOR >= 6)
    /// Retrieves the information for the given path
    ///
    /// \param path Absolute or relative path of the file system entry whose information is to be retrieved.
    /// \param flags Flags to moderate the search. Only the type flags (flags & RECLS_F_TYPEMASK) and RECLS_F_DIRECTORY_PARTS are recognised
    ///
    /// \return A valid instance if the path/flags exists, or a 'null' instance. (A null instance \c e evaluates \c !e to be true).
    static FileEntry Stat(char_type const *path, recls_uint32_t flags = RECLS_F_FILES | RECLS_F_DIRECTORIES | RECLS_F_DIRECTORY_PARTS);
#endif /* recls 1.6.x */
/// @}

// Implementation
private:
    static ctor_args_type CreateSearch_(char const *rootDir, char const *pattern, recls_uint32_t flags)
    {
        hrecls_t    hSrch;
        recls_rc_t  rc = Recls_Search(rootDir, pattern, flags, &hSrch);

        return Search::ctor_args_type(hSrch, rc);
    }

// Not to be implemented
private:
    FileSearch(class_type const &rhs);
    class_type &operator =(class_type const &rhs);
};

/* /////////////////////////////////////////////////////////////////////////////
 * Implementation
 */

inline FileSearch::FileSearch(FileSearch::char_type const *rootDir, FileSearch::char_type const *pattern, recls_uint32_t flags)
    : Search(CreateSearch_(rootDir, pattern, flags))
{}

inline FileSearch::~FileSearch()
{}

#if defined(RECLS_DOCUMENTATION_SKIP_SECTION) || \
    (   RECLS_VER_MAJOR > 1 || \
        RECLS_VER_MINOR >= 6)
inline /* static */ FileEntry FileSearch::Stat(FileSearch::char_type const *path, recls_uint32_t flags /* = RECLS_F_FILES | RECLS_F_DIRECTORIES | RECLS_F_DIRECTORY_PARTS */)
{
    recls_info_t    info;
    recls_rc_t      rc  =   Recls_Stat(path, flags, &info);

    return RECLS_SUCCEEDED(rc) ? make_entry_(info) : FileEntry();
}
#endif /* recls 1.6.x */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace cpp */
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !RECLS_INCL_RECLS_CPP_HPP_FILESEARCH */

/* ////////////////////////////////////////////////////////////////////////// */
