/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls/stl/search_sequence.hpp
 *
 * Purpose:     Contains the basic_search_sequence template class, and ANSI
 *              and Unicode specialisations thereof.
 *
 * Created:     10th September 2003
 * Updated:     19th May 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_RECLSTL_STL_HPP_SEARCH_SEQUENCE
#define RECLS_INCL_RECLSTL_STL_HPP_SEARCH_SEQUENCE

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_RECLSTL_STL_HPP_SEARCH_SEQUENCE_MAJOR    3
# define RECLS_VER_RECLSTL_STL_HPP_SEARCH_SEQUENCE_MINOR    1
# define RECLS_VER_RECLSTL_STL_HPP_SEARCH_SEQUENCE_REVISION 3
# define RECLS_VER_RECLSTL_STL_HPP_SEARCH_SEQUENCE_EDIT     38
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file recls/stl/search_sequence.hpp Definition of the \link recls::stl::basic_search_sequence basic_search_sequence\endlink 
 * class, and supporting classes 
 * (\link recls::stl::basic_search_sequence_const_iterator basic_search_sequence_const_iterator\endlink and 
 * \link recls::stl::basic_search_sequence_value_type basic_search_sequence_value_type\endlink), representing the
 * \ref group_recls_stl "recls STL mapping" */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef RECLS_INCL_RECLS_STL_HPP_RECLS
# include <recls/stl/recls.hpp>
#endif /* !RECLS_INCL_RECLS_STL_HPP_RECLS */
#ifndef RECLS_INCL_RECLS_STL_HPP_TRAITS
# include <recls/stl/traits.hpp>
#endif /* !RECLS_INCL_RECLS_STL_HPP_TRAITS */
#if defined(RECLS_PLATFORM_IS_UNIX)
# include <unixstl/file_path_buffer.hpp>
#elif defined(RECLS_PLATFORM_IS_WIN32)
# include <winstl/file_path_buffer.hpp>
#else
# error Platform not (yet) recognised
#endif /* platform */
#include <stlsoft/simple_string.hpp>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{

namespace stl
{
#endif /* !RECLS_NO_NAMESPACE */

#if defined(RECLS_PLATFORM_IS_WIN32)
typedef winstl::basic_file_path_buffer<recls_char_t>    file_path_buffer;
#elif defined(RECLS_PLATFORM_IS_UNIX)
typedef unixstl::basic_file_path_buffer<recls_char_t>   file_path_buffer;
#else
# error Platform not (yet) recognised
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION

template<   typename C
        ,   typename T
        >
class basic_search_sequence_value_type;

template<   typename C
        ,   typename T
        ,   typename V
        >
class basic_search_sequence_const_iterator;

#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Utility classes
 */

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
struct rss_shared_handle
{
    hrecls_t        hSrch;
    recls_sint32_t  cRefs;

public:
    explicit rss_shared_handle(hrecls_t h)
        : hSrch(h)
        , cRefs(1)
    {}
    void Release()
    {
        if(--cRefs == 0)
        {
            delete this;
        }
    }
#if defined(__STLSOFT_COMPILER_IS_GCC)
protected:
#else /* ? __STLSOFT_COMPILER_IS_GCC */
private:
#endif /* __STLSOFT_COMPILER_IS_GCC */
    ~rss_shared_handle()
    {
        recls_message_assert("Shared search handle being destroyed with outstanding references!", 0 == cRefs);

        if(NULL != hSrch)
        {
            Recls_SearchClose(hSrch);
        }
    }
};
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// class basic_search_sequence
/// Presents an STL-like sequence interface over the items on the file-system
///
/// \ingroup group_recls_stl
///
/// \param C The character type
/// \param T The traits type. On translators that support default template arguments this defaults to reclstl_traits<C>
template<   typename C
#ifdef __STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT
        ,   typename T = reclstl_traits<C>
#else
        ,   typename T /* = reclstl_traits<C> */
#endif /* __STLSOFT_CF_TEMPLATE_CLASS_DEFAULT_CLASS_ARGUMENT_SUPPORT */
        >
class basic_search_sequence
{
/// \name Types
/// @{
public:
    /// The character type
    typedef C                                                       char_type;
    /// The traits type
    typedef T                                                       traits_type;
    /// The current parameterisation of the type
    typedef basic_search_sequence<C, T>                             class_type;
    /// The value type
    typedef basic_search_sequence_value_type<C, T>                  value_type;
    /// The non-mutating (const) iterator type supporting the Input Iterator concept
    typedef basic_search_sequence_const_iterator<C, T, value_type>  const_iterator;
    /// The reference type
    typedef value_type                                              &reference;
    /// The non-mutable (const) reference type
    typedef value_type const                                        &const_reference;
    /// The find-data type
    typedef ss_typename_type_k traits_type::entry_type              entry_type;
    /// The size type
    typedef size_t                                                  size_type;
    /// The difference type
    typedef ptrdiff_t                                               difference_type;
    /// The string type
    typedef stlsoft::basic_simple_string<C>                         string_type;
/// @}

/// \name Construction
/// @{
public:
    /// Commence a search according to the given search pattern and flags
    basic_search_sequence(char_type const *pattern, recls_uint32_t flags);
#if defined(__STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT)
    /// Commence a search according to the given search pattern and flags
    template <typename S>
    basic_search_sequence(S const &pattern, recls_uint32_t flags)
        : m_directory(copy_or_null_(m_directory_, stlsoft_ns_qual(c_str_ptr)("")))
        , m_pattern(copy_or_null_(m_pattern_, stlsoft_ns_qual(c_str_ptr)(pattern)))
        , m_flags(flags)
    {}
#endif /* __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT */
    /// Commence a search according to the given search pattern and flags, relative to \c directory
    basic_search_sequence(char_type const *directory, char_type const *pattern, recls_uint32_t flags);
#if defined(__STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT)
    /// Commence a search according to the given search pattern and flags, relative to \c directory
    template <typename S1, typename S2>
    basic_search_sequence(S1 const &directory, S2 const &pattern, recls_uint32_t flags)
        : m_directory(copy_or_null_(m_directory_, stlsoft_ns_qual(c_str_ptr)(directory)))
        , m_pattern(copy_or_null_(m_pattern_, stlsoft_ns_qual(c_str_ptr)(pattern)))
        , m_flags(flags)
    {}
#endif /* __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT */
#if defined(RECLS_API_FTP)
    /// Commence an FTP search on the given site, according to the given search pattern and flags 
    basic_search_sequence(char_type const *host, char_type const *username, char_type const *password, char_type const *directory, char_type const *pattern, recls_uint32_t flags);
#endif /* RECLS_API_FTP */
    /// Destructor
    ~basic_search_sequence()
    {
        stlsoft_static_assert(stlsoft_raw_offsetof(class_type, m_directory_) < stlsoft_raw_offsetof(class_type, m_directory));
        stlsoft_static_assert(stlsoft_raw_offsetof(class_type, m_pattern_) < stlsoft_raw_offsetof(class_type, m_pattern));
    }
/// @}

/// \name Iteration
/// @{
public:
    /// Begins the iteration
    ///
    /// \return An iterator representing the start of the sequence
    const_iterator  begin() const;
    /// Ends the iteration
    ///
    /// \return An iterator representing the end of the sequence
    const_iterator  end() const;
/// @}

/// \name State
/// @{
public:
    /// Returns the number of items in the sequence
    size_type           size() const;
    /// Indicates whether the sequence is empty
    recls_bool_t        empty() const;
    /// Returns the maximum number of items in the sequence
    static size_type    max_size();
/// @}

/// \name Members
/// @{
private:
    friend class basic_search_sequence_value_type<C, T>;
    friend class basic_search_sequence_const_iterator<C, T, value_type>;

    // This one is implemented in-class as it allows sequence to be used by VC++ 5
    static char_type const  *copy_or_null_(file_path_buffer &dest, char_type const *src)
    {
        return (NULL == src) ? static_cast<char_type const*>(NULL) : traits_type::str_copy(&dest[0], src);
    }

#if defined(RECLS_API_FTP)
    string_type             m_host;
    string_type             m_username;
    string_type             m_password;
#endif /* RECLS_API_FTP */
    // TODO: Lose the file_path_buffer, and use auto_buffer directly
    file_path_buffer        m_directory_;
    file_path_buffer        m_pattern_;
    char_type const *const  m_directory;
    char_type const *const  m_pattern;
    recls_uint32_t          m_flags;
/// @}

// Not to be implemented
private:
    basic_search_sequence(class_type const &);
    basic_search_sequence const &operator =(class_type const &);
};

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs for commonly encountered types
 */

/// Instantiation of the basic_search_sequence template for the ANSI character type \c recls_char_a_t
typedef basic_search_sequence<recls_char_a_t, reclstl_traits<recls_char_a_t> >      search_sequence_a;
/// Instantiation of the basic_search_sequence template for the Unicode character type \c recls_char_w_t
typedef basic_search_sequence<recls_char_w_t, reclstl_traits<recls_char_w_t> >      search_sequence_w;

/* ////////////////////////////////////////////////////////////////////////// */

// class basic_search_sequence_value_type
/// Value type for the basic_search_sequence
///
/// \ingroup group_recls_stl
template<   typename C
        ,   typename T
        >
class basic_search_sequence_value_type
{
/// \name Types
/// @{
public:
    /// The character type
    typedef C                                                       char_type;
    /// The traits type
    typedef T                                                       traits_type;
    /// The current parameterisation of the type
    typedef basic_search_sequence_value_type<C, T>                  class_type;
    /// The find-data type
    typedef ss_typename_type_k traits_type::entry_type              entry_type;
    /// The directory parts sequence type
    typedef ss_typename_type_k traits_type::directory_parts_type    directory_parts_type;
/// @}

/// \name Construction
/// @{
public:
    /// Default constructor
    basic_search_sequence_value_type();
    basic_search_sequence_value_type(class_type const &rhs);
private:
    basic_search_sequence_value_type(entry_type info)
        : m_info(info)
    {}
public:
    /// Destructor
    ~basic_search_sequence_value_type();

    /// Copy assignment operator
    class_type &operator =(class_type const &rhs);
/// @}

/// \name Attributes
/// @{
public:
    /// Returns the full path of the item
    string_t                get_path() const;
#ifdef RECLS_PLATFORM_API_WIN32
    char_type               get_drive() const;
#endif /* RECLS_PLATFORM_API_WIN32 */
    /// Returns the directory of the item
    string_t                get_directory() const;
    /// Returns the directory (+ drive) of the item
    string_t                get_directory_path() const;
#if defined(RECLS_DOCUMENTATION_SKIP_SECTION) || \
    (   RECLS_VER_MAJOR > 1 || \
        RECLS_VER_MINOR >= 6)
    /// Returns the UNC drive component, if any, of the item's path
    string_t                get_UNC_drive() const;
#endif /* recls 1.6.x */
#ifndef RECLS_PURE_API
    /// Returns a sequence of directory parts of the item
    directory_parts_type    get_directory_parts() const;
#endif /* !RECLS_PURE_API */
    /// Returns the file (file name + extension) of the item
    string_t                get_file() const;
    /// Returns the short form of the filename of the item
    string_t                get_short_file() const;
    /// Returns the filename of the item
    string_t                get_filename() const;
    /// Returns the file extension of the item
    string_t                get_fileext() const;

    recls_time_t            get_creation_time() const;
    recls_time_t            get_modification_time() const;
    recls_time_t            get_last_access_time() const;
    recls_time_t            get_last_status_change_time() const;

    recls_filesize_t        get_size() const;

    recls_bool_t            is_readonly() const;
    recls_bool_t            is_directory() const;
    recls_bool_t            is_link() const;
#if defined(RECLS_DOCUMENTATION_SKIP_SECTION) || \
    (   RECLS_VER_MAJOR > 1 || \
        RECLS_VER_MINOR >= 6)
    /// Indicates whether the file entry's path is UNC
    recls_bool_t            is_UNC() const;
#endif /* recls 1.6.x */

#ifndef RECLS_PURE_API
    /// A C-string to the full path
    char_type const         *c_str() const;
#endif /* !RECLS_PURE_API */

    /// Returns a non-mutating reference to find-data
    entry_type const        &get_entry() const;
/// @}

/// Implementation
private:
#ifdef RECLS_PURE_API
    static string_t get_string_(size_t (RECLS_CALLCONV_DEFAULT *pfn)(recls_info_t, recls_char_t *, size_t), entry_type info)
    {
        file_path_buffer    buffer;
        size_t              cch =   pfn(info, &buffer[0], buffer.size());

        return string_t(&buffer[0], cch);
    }
#endif /* RECLS_PURE_API */

// Members
private:
    friend class basic_search_sequence_const_iterator<C, T, class_type>;

    entry_type  m_info;
};

// class basic_search_sequence_const_iterator
/// Iterator type for the basic_search_sequence supporting the Input Iterator concept
///
/// \ingroup group_recls_stl
template<   typename C
        ,   typename T
        ,   typename V
        >
class basic_search_sequence_const_iterator
    : public stlsoft_ns_qual(iterator_base)<stlsoft_ns_qual_std(input_iterator_tag), V, ptrdiff_t, void, V>
{
public:
    /// The character type
    typedef C                                               char_type;
    /// The traits type
    typedef T                                               traits_type;
    /// The value type
    typedef V                                               value_type;
    /// The current parameterisation of the type
    typedef basic_search_sequence_const_iterator<C, T, V>   class_type;
    /// The find-data type
    typedef ss_typename_type_k traits_type::entry_type      entry_type;
private:
    typedef basic_search_sequence<C, T>                     sequence_type;

private:
    explicit basic_search_sequence_const_iterator(hrecls_t hSrch)
        : m_handle(make_handle_(hSrch))
    {}
public:
    /// Default constructor
    basic_search_sequence_const_iterator();
    /// Copy constructor
    basic_search_sequence_const_iterator(class_type const &rhs);
    /// Destructor
    ~basic_search_sequence_const_iterator();

    // Copy assignment operator
    basic_search_sequence_const_iterator &operator =(class_type const &rhs);

public:
    /// Pre-increment operator
    class_type &operator ++();
    /// Post-increment operator
    void operator ++(int);
    /// Dereference to return the value at the current position
    const value_type operator *() const;
    /// Evaluates whether \c this and \c rhs are equivalent
    recls_bool_t operator ==(class_type const &rhs) const;
    /// Evaluates whether \c this and \c rhs are not equivalent
    recls_bool_t operator !=(class_type const &rhs) const;

// Implementation
private:
    rss_shared_handle   *make_handle_(hrecls_t hSrch);

// Members
private:
    friend class basic_search_sequence<C, T>;

    rss_shared_handle   *m_handle;
};

///////////////////////////////////////////////////////////////////////////////
// Shims

/// \brief is_empty shim
/// \ingroup group_recls_stl
///
/// This returns a non-zero value if the given sequence is empty.
/// \param s The sequence whose state is to be tested
///
/// \note This fits the <a href = "http://stlsoft.org/help/main.html">STLSoft</a> shims 
/// concept, described in <a href = "http://synesis.com.au/resources/articles/cpp/shims.pdf">this</a>
/// Synesis Software White Paper, and featured in the article "<a href = "http://www.cuj.com/documents/s=8681/cuj0308wilson/">Generalised String Manipulation: Access Shims and Type-tunnelling</a>",
/// in the August 2003 issue of <a href = "http://www.cuj.com">C/C++ User's Journal</a>.
template <typename C, typename T>
inline recls_bool_t is_empty(basic_search_sequence<C, T> const &s)
{
    return s.empty();
}

#ifndef RECLS_PURE_API
/// \brief c_str_ptr shim
/// \ingroup group_recls_stl
///
/// This returns a non-null null-terminated C-string of the given value's full path.
/// \param v The basic_search_sequence_value_type instance whose path is to be returned
///
/// \note This fits the <a href = "http://stlsoft.org/help/main.html">STLSoft</a> shims 
/// concept, described in <a href = "http://synesis.com.au/resources/articles/cpp/shims.pdf">this</a>
/// Synesis Software White Paper, and featured in the article "<a href = "http://www.cuj.com/documents/s=8681/cuj0308wilson/">Generalised String Manipulation: Access Shims and Type-tunnelling</a>",
/// in the August 2003 issue of <a href = "http://www.cuj.com">C/C++ User's Journal</a>.
template <typename C, typename T>
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::char_type const *c_str_ptr(basic_search_sequence_value_type<C, T> const &v)
#else
inline C const *c_str_ptr(basic_search_sequence_value_type<C, T> const &v)
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */
{
    return v.c_str();
}

template <typename C, typename T>
inline size_t c_str_len(basic_search_sequence_value_type<C, T> const &v)
{
    return stlsoft_ns_qual(c_str_len)(v.c_str());
}
#endif /* !RECLS_PURE_API */

///////////////////////////////////////////////////////////////////////////////
// IOStream compatibility

template<   typename S
        ,   typename C
        ,   typename T
        >
inline S &operator <<(S &s, basic_search_sequence_value_type<C, T> const &v)
{
#ifdef RECLS_PURE_API
    s << v.c_str();
#else /* ? RECLS_PURE_API */
    s << v.get_path();
#endif /* RECLS_PURE_API */

    return s;
}

///////////////////////////////////////////////////////////////////////////////
// Implementation

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION

// basic_search_sequence


// Construction
template <typename C, typename T>
inline basic_search_sequence<C, T>::basic_search_sequence(char_type const *pattern, recls_uint32_t flags)
    : m_directory(copy_or_null_(m_directory_, ""))
    , m_pattern(copy_or_null_(m_pattern_, pattern))
    , m_flags(flags)
{}

template <typename C, typename T>
inline basic_search_sequence<C, T>::basic_search_sequence(char_type const *directory, char_type const *pattern, recls_uint32_t flags)
    : m_directory(copy_or_null_(m_directory_, directory))
    , m_pattern(copy_or_null_(m_pattern_, pattern))
    , m_flags(flags)
{}

#if defined(RECLS_API_FTP)
template <typename C, typename T>
inline basic_search_sequence<C, T>::basic_search_sequence(char_type const *host, char_type const *username, char_type const *password, char_type const *directory, char_type const *pattern, recls_uint32_t flags)
    : m_host(host)
    , m_username(username)
    , m_password(password)
    , m_directory(copy_or_null_(m_directory_, directory))
    , m_pattern(copy_or_null_(m_pattern_, pattern))
    , m_flags(flags)
{
    recls_message_assert("Must specify a hostname if using FTP", (NULL == username && NULL == password) || NULL != host);
}
#endif /* RECLS_API_FTP */

// Iteration
template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence<C, T>::const_iterator basic_search_sequence<C, T>::begin() const
{
#if defined(RECLS_API_FTP)
    if(m_host.empty())
    {
#endif /* RECLS_API_FTP */
        return const_iterator(traits_type::Search(m_directory, m_pattern, m_flags));
#if defined(RECLS_API_FTP)
    }
    else
    {
        return const_iterator(traits_type::SearchFtp(m_host.c_str(), m_username.c_str(), m_password.c_str(), m_directory, m_pattern, m_flags));
    }
#endif /* RECLS_API_FTP */
}

template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence<C, T>::const_iterator basic_search_sequence<C, T>::end() const
{
    return const_iterator();
}

// State
template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence<C, T>::size_type basic_search_sequence<C, T>::size() const
{
    const_iterator  b   =   begin();
    const_iterator  e   =   end();
    size_type       c   =   0;

    for(; b != e; ++b)
    {
        ++c;
    }

    return c;
}

template <typename C, typename T>
inline recls_bool_t basic_search_sequence<C, T>::empty() const
{
    return begin() == end();
}

template <typename C, typename T>
inline /* static */ ss_typename_type_k basic_search_sequence<C, T>::size_type basic_search_sequence<C, T>::max_size()
{
    return static_cast<size_type>(-1);
}

// basic_search_sequence_value_type


template <typename C, typename T>
inline basic_search_sequence_value_type<C, T>::basic_search_sequence_value_type()
    : m_info(NULL)
{}

template <typename C, typename T>
inline basic_search_sequence_value_type<C, T>::basic_search_sequence_value_type(ss_typename_type_k basic_search_sequence_value_type<C, T>::class_type const &rhs)
    : m_info(traits_type::CopyDetails(rhs.m_info))
{}

template <typename C, typename T>
inline basic_search_sequence_value_type<C, T>::~basic_search_sequence_value_type()
{
    if(NULL != m_info)
    {
        traits_type::CloseDetails(m_info);
    }
}

template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::class_type &basic_search_sequence_value_type<C, T>::operator =(ss_typename_type_k basic_search_sequence_value_type<C, T>::class_type const &rhs)
{
    if(NULL != m_info)
    {
        traits_type::CloseDetails(m_info);
    }

    m_info = traits_type::CopyDetails(rhs.m_info);

    return *this;
}

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_path() const
{
    recls_assert(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetPathProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->path.begin, m_info->path.end);
#endif /* RECLS_PURE_API */
}

#ifdef RECLS_PLATFORM_API_WIN32
template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::char_type basic_search_sequence_value_type<C, T>::get_drive() const
{
    char_type   chDrive;

    return (Recls_GetDriveProperty(m_info, &chDrive), chDrive);
}
#endif /* RECLS_PLATFORM_API_WIN32 */

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_directory() const
{
    recls_assert(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetDirectoryProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->directory.begin, m_info->directory.end);
#endif /* RECLS_PURE_API */
}

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_directory_path() const
{
    recls_assert(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetDirectoryPathProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->path.begin, m_info->directory.end);
#endif /* RECLS_PURE_API */
}

#if defined(RECLS_DOCUMENTATION_SKIP_SECTION) || \
    (   RECLS_VER_MAJOR > 1 || \
        RECLS_VER_MINOR >= 6)
template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_UNC_drive() const
{
    recls_assert(NULL != m_info);

#if 0
    if(!is_UNC())
    {
        return string_t();
    }
    else
    {
#ifdef RECLS_PURE_API
        string_t    directoryPath   =   get_directory_path();
        size_t      directoryLen    =   Recls_GetDirectoryProperty(m_info, NULL, 0);

        return string_t(directoryPath, 0, directoryPath.length() - directoryLen);
#else /* ? RECLS_PURE_API */
        return string_t(m_info->path.begin, m_info->directory.begin);
#endif /* RECLS_PURE_API */
    }
#else /* ? 0 */
# ifdef RECLS_PURE_API
    return get_string_(Recls_GetUNCDriveProperty, m_info);
# else /* ? RECLS_PURE_API */
    return string_t(m_info->path.begin, m_info->directory.begin);
# endif /* RECLS_PURE_API */
#endif /* 0 */
}
#endif /* recls 1.6.x */

#ifndef RECLS_PURE_API
template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::directory_parts_type basic_search_sequence_value_type<C, T>::get_directory_parts() const
{
    recls_assert(NULL != m_info);

    return directory_parts_type(m_info->directoryParts.begin, m_info->directoryParts.end);
}
#endif /* !RECLS_PURE_API */

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_file() const
{
    recls_assert(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetFileProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->fileName.begin, m_info->fileExt.end);
#endif /* RECLS_PURE_API */
}

#if 0
template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_short_file() const
{
    return m_info.cAlternateFileName[0] != '\0' ? m_info.cAlternateFileName : m_info.cFileName;
}
#endif /* 0 */

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_filename() const
{
    recls_assert(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetFileNameProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->fileName.begin, m_info->fileName.end);
#endif /* RECLS_PURE_API */
}

template <typename C, typename T>
inline string_t basic_search_sequence_value_type<C, T>::get_fileext() const
{
    recls_assert(NULL != m_info);

#ifdef RECLS_PURE_API
    return get_string_(Recls_GetFileExtProperty, m_info);
#else /* ? RECLS_PURE_API */
    return string_t(m_info->fileExt.begin, m_info->fileExt.end);
#endif /* RECLS_PURE_API */
}

template <typename C, typename T>
inline recls_time_t basic_search_sequence_value_type<C, T>::get_creation_time() const
{
    recls_assert(NULL != m_info);

    return Recls_GetCreationTime(m_info);
}

template <typename C, typename T>
inline recls_time_t basic_search_sequence_value_type<C, T>::get_modification_time() const
{
    recls_assert(NULL != m_info);

    return Recls_GetModificationTime(m_info);
}

template <typename C, typename T>
inline recls_time_t basic_search_sequence_value_type<C, T>::get_last_access_time() const
{
    recls_assert(NULL != m_info);

    return Recls_GetLastAccessTime(m_info);
}

template <typename C, typename T>
inline recls_time_t basic_search_sequence_value_type<C, T>::get_last_status_change_time() const
{
    recls_assert(NULL != m_info);

    return Recls_GetLastStatusChangeTime(m_info);
}

template <typename C, typename T>
inline recls_filesize_t basic_search_sequence_value_type<C, T>::get_size() const
{
    recls_assert(NULL != m_info);

#ifdef RECLS_PURE_API
    recls_filesize_t    size;

    return (Recls_GetSizeProperty(m_info, &size), size);
#else /* ? RECLS_PURE_API */
    return m_info->size;
#endif /* RECLS_PURE_API */
}

template <typename C, typename T>
inline recls_bool_t basic_search_sequence_value_type<C, T>::is_readonly() const
{
    recls_assert(NULL != m_info);

    return Recls_IsFileReadOnly(m_info);
}

template <typename C, typename T>
inline recls_bool_t basic_search_sequence_value_type<C, T>::is_directory() const
{
    recls_assert(NULL != m_info);

    return Recls_IsFileDirectory(m_info);
}

template <typename C, typename T>
inline recls_bool_t basic_search_sequence_value_type<C, T>::is_link() const
{
    recls_assert(NULL != m_info);

    return Recls_IsFileLink(m_info);
}

#if defined(RECLS_DOCUMENTATION_SKIP_SECTION) || \
    (   RECLS_VER_MAJOR > 1 || \
        RECLS_VER_MINOR >= 6)
template <typename C, typename T>
inline recls_bool_t basic_search_sequence_value_type<C, T>::is_UNC() const
{
    recls_assert(NULL != m_info);

    return Recls_IsFileUNC(m_info);
}
#endif /* recls 1.6.x */

#ifndef RECLS_PURE_API
template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::char_type const * basic_search_sequence_value_type<C, T>::c_str() const
{
    recls_assert(NULL != m_info);

    return m_info->path.begin;
}
#endif /* !RECLS_PURE_API */

template <typename C, typename T>
inline ss_typename_type_k basic_search_sequence_value_type<C, T>::entry_type const &basic_search_sequence_value_type<C, T>::get_entry() const
{
    return m_info;
}

// basic_search_sequence_const_iterator

template <typename C, typename T, typename V>
inline rss_shared_handle *basic_search_sequence_const_iterator<C, T, V>::make_handle_(hrecls_t hSrch)
{
    return (NULL != hSrch) ? new rss_shared_handle(hSrch) : static_cast<rss_shared_handle*>(NULL);
}

template <typename C, typename T, typename V>
inline basic_search_sequence_const_iterator<C, T, V>::basic_search_sequence_const_iterator()
    : m_handle(NULL)
{}

template <typename C, typename T, typename V>
inline basic_search_sequence_const_iterator<C, T, V>::basic_search_sequence_const_iterator(class_type const &rhs)
    : m_handle(rhs.m_handle)
{
    if(NULL != m_handle)
    {
        ++m_handle->cRefs;
    }
}

template <typename C, typename T, typename V>
inline ss_typename_type_k basic_search_sequence_const_iterator<C, T, V>::class_type &basic_search_sequence_const_iterator<C, T, V>::operator =(typename basic_search_sequence_const_iterator<C, T, V>::class_type const &rhs)
{
    if(NULL != m_handle)
    {
        m_handle->Release();
    }

    m_handle =   rhs.m_handle;

    if(NULL != m_handle)
    {
        ++m_handle->cRefs;
    }

    return *this;
}

template <typename C, typename T, typename V>
inline basic_search_sequence_const_iterator<C, T, V>::~basic_search_sequence_const_iterator()
{
    if(NULL != m_handle)
    {
        m_handle->Release();
    }
}

template <typename C, typename T, typename V>
inline ss_typename_type_k basic_search_sequence_const_iterator<C, T, V>::class_type &basic_search_sequence_const_iterator<C, T, V>::operator ++()
{
    recls_message_assert("Attempting to increment invalid iterator", NULL != m_handle);

    if(RECLS_FAILED(Recls_GetNext(m_handle->hSrch)))
    {
        m_handle->Release();

        m_handle = NULL;
    }

    return *this;
}

template <typename C, typename T, typename V>
inline void basic_search_sequence_const_iterator<C, T, V>::operator ++(int)
{
    operator ++();
}

template <typename C, typename T, typename V>
inline const ss_typename_type_k basic_search_sequence_const_iterator<C, T, V>::value_type basic_search_sequence_const_iterator<C, T, V>::operator *() const
{
    entry_type  info;

    recls_message_assert("Attempting to dereference invalid iterator", NULL != m_handle);

    if( m_handle->hSrch != NULL &&
        RECLS_SUCCEEDED(traits_type::GetDetails(m_handle->hSrch, &info)))
    {
        return value_type(info);
    }
    else
    {
        recls_message_assert("Dereferencing end()-valued iterator", 0);

        return value_type();
    }
}

template <typename C, typename T, typename V>
inline recls_bool_t basic_search_sequence_const_iterator<C, T, V>::operator ==(class_type const &rhs) const
{
    return (m_handle == NULL) && (rhs.m_handle == NULL);
}

template <typename C, typename T, typename V>
inline recls_bool_t basic_search_sequence_const_iterator<C, T, V>::operator !=(class_type const &rhs) const
{
    return ! operator ==(rhs);
}

#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////// */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace stl */
} /* namespace recls */

// Now we introduce the two shims into the STLSoft namespace - stlsoft.
namespace stlsoft
{
    using recls::stl::is_empty;
#ifndef RECLS_PURE_API
    using recls::stl::c_str_ptr;
    using recls::stl::c_str_len;
#endif /* !RECLS_PURE_API */

} // namespace stlsoft
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* RECLS_INCL_RECLSTL_STL_HPP_SEARCH_SEQUENCE */

/* ////////////////////////////////////////////////////////////////////////// */
