/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_FileSearch.cpp
 *
 * Purpose:     JNI implementation file for the FileSearch class
 *
 * Created      21st November 2003
 * Updated:     20th May 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#include <recls_FileSearch.h>

#include <recls.h>
#if defined(RECLS_PLATFORM_IS_WIN32)
# include <recls_win32.h>
#elif defined(RECLS_PLATFORM_IS_UNIX)
# include <recls_unix.h>
#else
# error Platform not (yet) recognised
#endif /* platform */

#include "recls_jni_util.h"

#include <stdarg.h>

/* ////////////////////////////////////////////////////////////////////////// */

#if !defined(RECLS_NO_NAMESPACE)
using recls::hrecls_t;
using recls::recls_info_t;
using recls::recls_strptrs_t;

using recls::Recls_Search;
using recls::Recls_GetDirectoryPartProperty;
using recls::Recls_Stat;
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/*
 * Class:     org_recls_FileSearch
 * Method:    Stat
 * Signature: (Ljava/lang/String;I)Lorg/recls/Entry;
 */
JNIEXPORT jobject JNICALL Java_org_recls_FileSearch_Stat
  (JNIEnv *henv, jobject obj, jstring jsPath, jint flags)
{
    recls_char_t const	*path	=   SafeStringUTFChars(henv, jsPath, NULL);
	recls_info_t		hEntry;
	recls_rc_t			rc		=	Recls_Stat(path, flags, &hEntry);
    jobject				entry;

    if(RECLS_FAILED(rc))
    {
        ThrowReclsException(henv, "Search failed", rc);

		entry = NULL;
    }
	else
	{
        jclass          clsEntry        =   henv->FindClass(CLASSNAME_ENTRY);
        jclass          clsString       =   henv->FindClass("Ljava/lang/String;");
        jmethodID       midEntryCtor    =   henv->GetMethodID(clsEntry, "<init>", "(ILjava/lang/String;[Ljava/lang/String;)V");
        size_t          cDirParts       =   Recls_GetDirectoryPartProperty(hEntry, -1, NULL, 0);
        jstring         path            =   StringFromRange(henv, hEntry->path.begin, hEntry->path.end);
        jobjectArray    dirParts        =   henv->NewObjectArray(cDirParts, clsString, NULL);

        for(size_t i = 0; i < cDirParts; ++i)
        {
            recls_strptrs_t const   *part   =   hEntry->directoryParts.begin + i;

            jstring dirPart = StringFromRange(henv, part->begin, part->end);

            henv->SetObjectArrayElement(dirParts, i, dirPart);
        }

        entry = henv->NewObject(clsEntry, midEntryCtor, hEntry, path, dirParts);
	}

	return entry;
}

/*
 * Class:     org_recls_FileSearch
 * Method:    initialise
 * Signature: (Ljava/lang/String;Ljava/lang/String;I)I
 */
JNIEXPORT jint JNICALL Java_org_recls_FileSearch_initialise
  (JNIEnv *henv, jclass cls, jstring jsSearchRoot, jstring jsPattern, jint flags)
{
    // Need to 
    recls_char_t const			*searchRoot		=   SafeStringUTFChars(henv, jsSearchRoot, NULL);
    recls_char_t const			*pattern		=   SafeStringUTFChars(henv, jsPattern, NULL);

    static const recls_char_t	searchRoot_[2]	=	{ '.', '\0' };
#if defined(RECLS_PLATFORM_IS_WIN32)
    static const recls_char_t	pattern_[4]		=	{ '*', '.', '*', '\0' };
#elif defined(RECLS_PLATFORM_IS_UNIX)
    static const recls_char_t	pattern_[2]		=	{ '*', '\0' };
#else
# error Platform not (yet) recognised
#endif /* platform */

    if(0 == *searchRoot)
    {
        searchRoot = searchRoot_;
    }
    if(0 == *pattern)
    {
        pattern = pattern_;
    }

    hrecls_t    hSrch;
    recls_rc_t  rc  =   Recls_Search(searchRoot, pattern, flags, &hSrch);

    if( RECLS_FAILED(rc) &&
        rc != RECLS_RC_NO_MORE_DATA)
    {
        ThrowReclsException(henv, "Search failed", rc);
    }

    return reinterpret_cast<jint>(hSrch);
}

/* ////////////////////////////////////////////////////////////////////////// */
