/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_Search.cpp
 *
 * Purpose:     JNI implementation file for the Search class
 *
 * Created      21st November 2003
 * Updated:     20th May 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#include <recls_Search.h>

#include <recls.h>
#if defined(RECLS_PLATFORM_IS_WIN32)
# include <recls_win32.h>
#elif defined(RECLS_PLATFORM_IS_UNIX)
# include <recls_unix.h>
#else
# error Platform not (yet) recognised
#endif /* platform */

#include "recls_jni_util.h"

#include <stdarg.h>

/* ////////////////////////////////////////////////////////////////////////// */

#if !defined(RECLS_NO_NAMESPACE)
using recls::hrecls_t;
using recls::recls_info_t;
using recls::recls_strptrs_t;

using recls::Recls_SearchClose;
using recls::Recls_GetLastError;
using recls::Recls_GetDetails;
using recls::Recls_GetNext;
using recls::Recls_GetDirectoryPartProperty;
using recls::Recls_GetRoots;
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

jfieldID GetSearchFieldID(JNIEnv *henv, jobject obj);
hrecls_t SearchFromJObject(JNIEnv *henv, jobject obj);

/* ////////////////////////////////////////////////////////////////////////// */


/*
 * Class:     org_recls_Search
 * Method:    getRoots
 * Signature: ()[Ljava/lang/String;
 */
JNIEXPORT jobjectArray JNICALL Java_org_recls_Search_getRoots
  (JNIEnv *henv, jclass cls)
{
    recls_root_t    roots[26];
    size_t          cRoots      =   Recls_GetRoots(roots, stlsoft_num_elements(roots));
    size_t          i;
    jclass          clsString   =   henv->FindClass("Ljava/lang/String;");
    jobjectArray    dirParts    =   henv->NewObjectArray(cRoots, clsString, NULL);

    for(i = 0; i < cRoots; ++i)
    {
#ifdef RECLS_CHAR_TYPE_IS_CHAR
        jstring     dirPart     =   henv->NewStringUTF(roots[i].name);
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR
        jstring     dirPart     =   henv->NewString(roots[i].name, wcslen(roots[i].name));
#else /* ?char type */
# error recls char type must be discriminated
#endif /* char type */
        henv->SetObjectArrayElement(dirParts, i, dirPart);
    }

    return dirParts;
}


/*
 * Class:     org_recls_Search
 * Method:    close
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_org_recls_Search_close
  (JNIEnv *henv, jobject obj)
{
    // Get the handle from the object
    hrecls_t    hSrch   =   SearchFromJObject(henv, obj);

    if(NULL == hSrch)
    {
//      ThrowException(henv, CLASSNAME_ILLEGALSTATEX, "Entry is already closed");
    }
    else
    {
        jfieldID    fid         =   GetSearchFieldID(henv, obj);

        henv->SetIntField(obj, fid, 0);

        Recls_SearchClose(hSrch);
    }
}

/*
 * Class:     org_recls_Search
 * Method:    hasMoreElements_
 * Signature: ()Z
 */
JNIEXPORT jboolean JNICALL Java_org_recls_Search_hasMoreElements_1
  (JNIEnv *henv, jobject obj)
{
    hrecls_t    hSrch   =   SearchFromJObject(henv, obj);

    if(NULL == hSrch)
    {
        return JNI_FALSE;
    }
    else
    {
        recls_rc_t  rc  =   Recls_GetLastError(hSrch);

        return RECLS_SUCCEEDED(rc) ? JNI_TRUE : JNI_FALSE;
    }
}

/*
 * Class:     org_recls_Search
 * Method:    nextElement_
 * Signature: ()Ljava/lang/Object;
 */
JNIEXPORT jobject JNICALL Java_org_recls_Search_nextElement_1
  (JNIEnv *henv, jobject obj)
{
    // This must return the current one, and then advance to the next
    hrecls_t    hSrch   =   SearchFromJObject(henv, obj);
    recls_rc_t  rc;
    jobject     entry;

    if( NULL == hSrch ||
        RECLS_FAILED(rc = Recls_GetLastError(hSrch)))
    {
        ThrowException(henv, CLASSNAME_NOSUCHELEMENTX, "No more elements in search");

        entry = NULL;
    }
    else
    {
        recls_info_t    hEntry;

        rc = Recls_GetDetails(hSrch, &hEntry);

        if(RECLS_SUCCEEDED(rc))
        {
            jclass          clsEntry        =   henv->FindClass(CLASSNAME_ENTRY);
            jclass          clsString       =   henv->FindClass("Ljava/lang/String;");
            jmethodID       midEntryCtor    =   henv->GetMethodID(clsEntry, "<init>", "(ILjava/lang/String;[Ljava/lang/String;)V");
            size_t          cDirParts       =   Recls_GetDirectoryPartProperty(hEntry, -1, NULL, 0);
            jstring         path            =   StringFromRange(henv, hEntry->path.begin, hEntry->path.end);
            jobjectArray    dirParts        =   henv->NewObjectArray(cDirParts, clsString, NULL);



            for(size_t i = 0; i < cDirParts; ++i)
            {
                recls_strptrs_t const   *part   =   hEntry->directoryParts.begin + i;

                jstring dirPart = StringFromRange(henv, part->begin, part->end);

                henv->SetObjectArrayElement(dirParts, i, dirPart);
            }

            entry = henv->NewObject(clsEntry, midEntryCtor, hEntry, path, dirParts);

            // (ii) Now do a get-next, and if that returns NO_MORE_DATA, close the element
            rc = Recls_GetNext(hSrch);

            if(RECLS_FAILED(rc))
            {
                jfieldID    fid = GetSearchFieldID(henv, obj);

                henv->SetIntField(obj, fid, 0);

                Recls_SearchClose(hSrch);
            }
        }
        else
        {
            ThrowException(henv, CLASSNAME_ILLEGALSTATEX, "Search entry retrieval failed");
        }
    }

    return entry;   
}

/* ////////////////////////////////////////////////////////////////////////// */

jfieldID GetSearchFieldID(JNIEnv *henv, jobject obj)
{
    jclass      clsEntry    =   henv->GetObjectClass(obj);
    jfieldID    fid         =   henv->GetFieldID(clsEntry, FIELDNAME_HSRCH, "I");

//  henv->DeleteLocalRef(clsEntry);

    return fid;
}

hrecls_t SearchFromJObject(JNIEnv *henv, jobject obj)
{
    jfieldID    fid = GetSearchFieldID(henv, obj);

    return reinterpret_cast<hrecls_t>(henv->GetIntField(obj, fid));
}

/* ////////////////////////////////////////////////////////////////////////// */


