/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_api.cpp
 *
 * Purpose:     Main (platform-independent) implementation file for the recls API.
 *
 * Created:     16th August 2003
 * Updated:     20th May 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls.h>
#include <recls_assert.h>
#include "recls_impl.h"
#include "recls_util.h"     // for recls_strlen(), recls_strncpy()

#include <stlsoft.h>
#if !defined(_STLSOFT_VER_1_8_3)
# error STLSoft version 1.8.3 or greater required. (www.stlsoft.org/downloads.html)
#endif /* STLSoft version */
#if !defined(RECLS_COMPILER_IS_VECTORC)
# include <stlsoft_nulldef.h>
#endif /* compiler */

#if defined(RECLS_PLATFORM_IS_UNIX)

# include <unixstl/path.hpp>

namespace platform_stl = ::unixstl;

#elif defined(RECLS_PLATFORM_IS_WIN32)

# include <winstl/path.hpp>

namespace platform_stl = ::winstl;

#else /* unrecognised platform */

# error The platform is not recognised. Only RECLS_PLATFORM_IS_UNIX or RECLS_PLATFORM_IS_WIN32 currently supported

#endif /* platform */


#include "recls_debug.h"

#include "ReclsSearch.hpp"
#include "ReclsFileSearch.hpp"
#include "EntryFunctions.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

namespace {

typedef ::platform_stl::filesystem_traits<recls_char_t>         traits_type;
typedef ::platform_stl::basic_file_path_buffer<recls_char_t>    file_path_buffer_type;
typedef ::platform_stl::basic_path<recls_char_t>                path_type;

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */

static const recls_char_t   s_defaultSearchRoot[]       =   RECLS_LITERAL(".");

#if defined(RECLS_PLATFORM_IS_UNIX)
# define _GetCreationTime           modificationTime
# define _GetLastStatusChangeTime   lastStatusChangeTime
# define RAPI_PATHNAMESEP           '/'
# define RAPI_PATHSEP               ':'
# define RAPI_WILDCARDSALL          '*'
#elif defined(RECLS_PLATFORM_IS_WIN32)
# define _GetCreationTime           creationTime
# define _GetLastStatusChangeTime   modificationTime
# define RAPI_PATHNAMESEP           '\\'
# define RAPI_PATHSEP               ';'
# define RAPI_WILDCARDSALL          '*', '.', '*'
#else /* unrecognised platform */
# error platform is not recognised
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Search control
 */

RECLS_FNDECL(recls_rc_t) Recls_Search(  recls_char_t const  *searchRoot
                                    ,   recls_char_t const  *pattern
                                    ,   recls_uint32_t      flags
                                    ,   hrecls_t            *phSrch)
{
    function_scope_trace("Recls_Search");

    recls_assert(NULL != phSrch);

    *phSrch = static_cast<hrecls_t>(0);

    recls_rc_t  rc;

#ifdef _RECLS_EXCEPTION_SUPPORT
    try
    {
#endif /* _RECLS_EXCEPTION_SUPPORT */
        // Default the input parameters

        // Default the search root
        if( NULL == searchRoot ||
            0 == *searchRoot)
        {
            searchRoot = s_defaultSearchRoot;
        }

        // Default the pattern
        if(NULL == pattern)
        {
            pattern = traits_type::pattern_all();
        }

        // Cater for the situation whereby the pattern is a file
        path_type               path(pattern);
        file_path_buffer_type   searchRoot_;

        if(/*  NULL == traits_type::str_str(pattern, Recls_GetPathSeparator()) &&
             */path.exists())
        {
            path.make_absolute(true);
#ifdef _RECLS_EXCEPTION_SUPPORT
            path.canonicalise(true);
#endif /* _RECLS_EXCEPTION_SUPPORT */

            recls_char_t const  *file   =   path.get_file();
            const size_t        cch     =   file - path.c_str();

            traits_type::str_n_copy(&searchRoot_[0], path.c_str(), cch);
            searchRoot_[cch] = '\0';
            searchRoot = searchRoot_.c_str();

            pattern = file;

            if(0 == (flags & RECLS_F_TYPEMASK))
            {
                if(traits_type::is_directory(path.c_str()))
                {
                    flags |= RECLS_F_DIRECTORIES;
                }
                else
                {
                    flags |= RECLS_F_FILES;
                }
            }
        }

        // Default the flags
        if(0 == (flags & RECLS_F_TYPEMASK))
        {
            flags |= RECLS_F_FILES;
        }

        if(0 == (flags & (RECLS_F_FILES | RECLS_F_DIRECTORIES)))
        {
            rc = RECLS_RC_INVALID_SEARCH_TYPE;
        }
        // Validate the pattern.
        else
        {
            rc = IsValidPattern(pattern, flags);
        }

        if(RECLS_SUCCEEDED(rc))
        {
            ReclsFileSearch *si;

            rc = ReclsFileSearch::FindAndCreate(searchRoot, pattern, flags, &si);

            if(RECLS_SUCCEEDED(rc))
            {
                *phSrch = ReclsSearch::ToHandle(si);

                rc = RECLS_RC_OK;
            }
        }
#ifdef _RECLS_EXCEPTION_SUPPORT
    }
    catch(std::exception &x)
    {
        debug_printf("Exception in Recls_Search(): %s\n", x.what());

        rc = RECLS_RC_UNEXPECTED;
    }
#endif /* _RECLS_EXCEPTION_SUPPORT */

    recls_message_assert("Must not return a success code with a null search handle", RECLS_RC_OK != rc || NULL != *phSrch);

    return rc;
}

/** Closes the given search */
RECLS_FNDECL(void) Recls_SearchClose(hrecls_t hSrch)
{
    function_scope_trace("Recls_SearchClose");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    recls_message_assert("Search handle is null!", NULL != si);

    delete si;
}

RECLS_FNDECL(recls_rc_t) Recls_Stat(    recls_char_t const  *path
                                    ,   recls_uint32_t      flags
                                    ,   recls_info_t        *phEntry)
{
    function_scope_trace("Recls_SearchClose");

    recls_assert(NULL != path);
    recls_assert(NULL != phEntry);

    *phEntry = static_cast<recls_info_t>(0);

#ifdef _RECLS_EXCEPTION_SUPPORT
    try
    {
#endif /* _RECLS_EXCEPTION_SUPPORT */
        path_type   path_(path);

        if(!path_.exists())
        {
            recls_char_t const  *file   =   path_.get_file();
            path_type           path2_(path_.c_str(), file - path_.c_str());

            if(!path2_.exists())
            {
                return RECLS_RC_INVALID_DIRECTORY;
            }

            return RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            if(traits_type::is_directory(path_.c_str()))
            {
                if(RECLS_F_FILES == (flags & (RECLS_F_FILES | RECLS_F_DIRECTORIES)))
                {
                    return RECLS_RC_WRONG_ENTRY_TYPE;
                }
            }
            else
            {
                if(RECLS_F_DIRECTORIES == (flags & (RECLS_F_FILES | RECLS_F_DIRECTORIES)))
                {
                    return RECLS_RC_WRONG_ENTRY_TYPE;
                }
            }

            path_.make_absolute(false);
#ifdef _RECLS_EXCEPTION_SUPPORT
            path_.canonicalise(false);
#endif /* _RECLS_EXCEPTION_SUPPORT */

            return ReclsFileSearch::Stat(path_.c_str(), flags, phEntry);
        }
#ifdef _RECLS_EXCEPTION_SUPPORT
    }
    catch(std::exception &x)
    {
        debug_printf("Exception in Recls_Stat(): %s\n", x.what());

        return RECLS_RC_UNEXPECTED;
    }
#endif /* _RECLS_EXCEPTION_SUPPORT */
}

/* /////////////////////////////////////////////////////////////////////////////
 * Search enumeration
 */

RECLS_FNDECL(recls_rc_t) Recls_SearchProcess(   recls_char_t const          *searchRoot
                                            ,   recls_char_t const          *pattern
                                            ,   recls_uint32_t              flags
                                            ,   hrecls_process_fn_t         pfn
                                            ,   recls_process_fn_param_t    param)
{
    function_scope_trace("Recls_SearchProcess");

    recls_assert(NULL != pfn);

    hrecls_t    hSrch;
    recls_rc_t  rc  =   Recls_Search(searchRoot, pattern, flags, &hSrch);

    if(RECLS_SUCCEEDED(rc))
    {
        recls_info_t    info;

        do
        {
            rc = Recls_GetDetails(hSrch, &info);

            if(RECLS_FAILED(rc))
            {
                break;
            }
            else
            {
                int res =   (*pfn)(info, param);

                Recls_CloseDetails(info);

                if(0 == res)
                {
                    rc = RECLS_RC_SEARCH_CANCELLED;

                    break;
                }
            }
        }
        while(RECLS_SUCCEEDED(rc = Recls_GetNext(hSrch)));

        Recls_SearchClose(hSrch);
    }

    if(RECLS_RC_NO_MORE_DATA == rc)
    {
        rc = RECLS_RC_OK;
    }

    return rc;
}

RECLS_FNDECL(recls_rc_t) Recls_GetNext(hrecls_t hSrch)
{
    function_scope_trace("Recls_GetNext");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    recls_message_assert("Search handle is null!", NULL != si);

    return si->GetNext();
}

RECLS_FNDECL(recls_rc_t) Recls_GetDetails(  hrecls_t        hSrch
                                        ,   recls_info_t    *pinfo)
{
    function_scope_trace("Recls_GetDetails");

    ReclsSearch     *si     =   ReclsSearch::FromHandle(hSrch);

    recls_message_assert("Search handle is null!", NULL != si);
    recls_assert(NULL != pinfo);

    return si->GetDetails(pinfo);
}

RECLS_FNDECL(recls_rc_t) Recls_GetNextDetails(  hrecls_t        hSrch
                                            ,   recls_info_t    *pinfo)
{
    function_scope_trace("Recls_GetNextDetails");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    recls_message_assert("Search handle is null!", NULL != si);

    return si->GetNextDetails(pinfo);
}

/***************************************
 * File entry info structure
 */

RECLS_FNDECL(void) Recls_CloseDetails(recls_info_t fileInfo)
{
    function_scope_trace("Recls_CloseDetails");

    recls_assert(NULL != fileInfo);

    Entry_Release(fileInfo);
}

RECLS_FNDECL(recls_rc_t) Recls_CopyDetails( recls_info_t    fileInfo
                                        ,   recls_info_t    *pinfo)
{
    function_scope_trace("Recls_CopyDetails");

    recls_assert(NULL != fileInfo);

    return Entry_Copy(fileInfo, pinfo);
}

RECLS_FNDECL(recls_rc_t) Recls_OutstandingDetails(  hrecls_t        hSrch
                                                ,   recls_uint32_t  *count)
{
    function_scope_trace("Recls_OutstandingDetails");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    recls_message_assert("Search handle is null!", NULL != si);
    recls_assert(NULL != count);
    STLSOFT_SUPPRESS_UNUSED(si);

    recls_sint32_t  cCreated;
    recls_sint32_t  cShared;

    Entry_BlockCount(&cCreated, &cShared);

    *count = cCreated;

    return RECLS_RC_OK;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Error handling
 */

RECLS_FNDECL(recls_rc_t) Recls_GetLastError(hrecls_t hSrch)
{
    function_scope_trace("Recls_GetLastError");

    ReclsSearch *si =   ReclsSearch::FromHandle(hSrch);

    recls_message_assert("Search handle is null!", NULL != si);

    return si->GetLastError();
}

static recls_char_t const *lookup_error_string_(recls_rc_t rc)
{
#if 1
    struct StringEntry
    {
        recls_rc_t          rc;
        recls_char_t const  *str;
    };

    static const StringEntry    entries[] = 
    {
            {   RECLS_RC_SEARCH_NO_CURRENT,         RECLS_LITERAL("Search has no current node")                                         }
        ,   {   RECLS_RC_INVALID_DIRECTORY,         RECLS_LITERAL("Invalid directory")                                                  }
        ,   {   RECLS_RC_NO_MORE_DATA,              RECLS_LITERAL("No more data")                                                       }
        ,   {   RECLS_RC_OUT_OF_MEMORY,             RECLS_LITERAL("No more memory")                                                     }
        ,   {   RECLS_RC_NOT_IMPLEMENTED,           RECLS_LITERAL("Not implemented")                                                    }
        ,   {   RECLS_RC_INVALID_SEARCH_TYPE,       RECLS_LITERAL("The search type was invalid")                                        }
        ,   {   RECLS_RC_INVALID_SEARCH_PROTOCOL,   RECLS_LITERAL("The search protocol was invalid")                                    }
        ,   {   RECLS_RC_UNEXPECTED,                RECLS_LITERAL("An unexpected internal condition was encountered")                   }
        ,   {   RECLS_RC_DOT_RECURSIVE_SEARCH,      RECLS_LITERAL("\"..\" specifed, or \".\" specified pattern in recursive search")    }
        ,   {   RECLS_RC_SEARCH_CANCELLED,          RECLS_LITERAL("The search was cancelled; it did not otherwise fail")                }
        ,   {   RECLS_RC_FTP_INIT_FAILED,           RECLS_LITERAL("FTP infrastructure initialisation failed")                           }
        ,   {   RECLS_RC_FTP_CONNECTION_FAILED,     RECLS_LITERAL("FTP connection failed/rejected")                                     }
        ,   {   RECLS_RC_WRONG_ENTRY_TYPE,          RECLS_LITERAL("Given entry was of wrong type")                                      }
        ,   {   RECLS_RC_CANNOT_STAT_ROOT,          RECLS_LITERAL("Cannot call stat() on root")                                         }
    };

    for(size_t i = 0; i < stlsoft_num_elements_(entries); ++i)
    {
        if(entries[i].rc == rc)
        {
            return entries[i].str;
        }
    }

    return RECLS_LITERAL("");
#else /* ? 0 */
    switch(rc)
    {
        case    RECLS_RC_SEARCH_NO_CURRENT:
            return RECLS_LITERAL("Search has no current node");
        
        case    RECLS_RC_INVALID_DIRECTORY:
            return RECLS_LITERAL("Invalid directory");
        
        case    RECLS_RC_NO_MORE_DATA:
            return RECLS_LITERAL("No more data");
        
        case    RECLS_RC_OUT_OF_MEMORY:
            return RECLS_LITERAL("No more memory");

        case    RECLS_RC_NOT_IMPLEMENTED:
            return RECLS_LITERAL("Not implemented");

        case    RECLS_RC_INVALID_SEARCH_TYPE:
            return RECLS_LITERAL("The search type was invalid");

        case    RECLS_RC_INVALID_SEARCH_PROTOCOL:
            return RECLS_LITERAL("The search protocol was invalid");

        case    RECLS_RC_UNEXPECTED:
            return RECLS_LITERAL("An unexpected internal condition was encountered");

        case    RECLS_RC_DOT_RECURSIVE_SEARCH:
            return RECLS_LITERAL("\"..\" specifed, or \".\" specified pattern in recursive search");

        case    RECLS_RC_SEARCH_CANCELLED:
            return RECLS_LITERAL("The search was cancelled; it did not otherwise fail");

        case    RECLS_RC_FTP_INIT_FAILED:
            return RECLS_LITERAL("FTP infrastructure initialisation failed");

        case    RECLS_RC_FTP_CONNECTION_FAILED:
            return RECLS_LITERAL("FTP connection failed/rejected");

        case    RECLS_RC_WRONG_ENTRY_TYPE:
            return RECLS_LITERAL("Given entry was of wrong type");

        case    RECLS_RC_CANNOT_STAT_ROOT:
            return RECLS_LITERAL("Cannot call stat() on root");

        default:
            return RECLS_LITERAL("");
    }
#endif /* 0 */
}

RECLS_FNDECL(size_t) Recls_GetErrorString(  recls_rc_t      rc
                                        ,   recls_char_t    *buffer
                                        ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetErrorString");

    size_t  cchError    =   recls_strlen(lookup_error_string_(rc));

    if(NULL == buffer)
    {
        return cchError;
    }
    else
    {
        return recls_strncpy(buffer, cchBuffer, lookup_error_string_(rc), cchError);
    }
}

RECLS_FNDECL(size_t) Recls_GetLastErrorString(  hrecls_t        hSrch
                                            ,   recls_char_t    *buffer
                                            ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetLastErrorString");

    return Recls_GetErrorString(Recls_GetLastError(hSrch), buffer, cchBuffer);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Property elicitation
 */

RECLS_FNDECL(size_t) Recls_GetPathProperty( recls_info_t    fileInfo
                                        ,   recls_char_t    *buffer
                                        ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetPathProperty");

    recls_assert(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->path, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetDirectoryProperty(    recls_info_t    fileInfo
                                                ,   recls_char_t    *buffer
                                                ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetDirectoryProperty");

    recls_assert(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->directory, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetDirectoryPathProperty(    recls_info_t    fileInfo
                                                    ,   recls_char_t    *buffer
                                                    ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetDirectoryPathProperty");

    recls_assert(NULL != fileInfo);

    struct recls_strptrs_t directoryPath =
    {
            fileInfo->path.begin    /* Directory path is defined by start of path ... */
        ,   fileInfo->directory.end /* ... to end of directory. */
    };

    return Recls_GetStringProperty_(&directoryPath, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetUNCDriveProperty( recls_info_t    fileInfo
                                            ,   recls_char_t    *buffer
                                            ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetUNCDriveProperty");

    recls_assert(NULL != fileInfo);

    struct recls_strptrs_t uncDrive =
    {
            fileInfo->path.begin        /* Directory path is defined by start of path ... */
        ,   fileInfo->directory.begin   /* ... to start of directory. */
    };

    if(!Recls_IsFileUNC(fileInfo))
    {
        uncDrive.end    =   uncDrive.begin;
    }

    return Recls_GetStringProperty_(&uncDrive, buffer, cchBuffer);
}


RECLS_FNDECL(size_t) Recls_GetFileProperty( recls_info_t    fileInfo
                                        ,   recls_char_t    *buffer
                                        ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetFileProperty");

    recls_assert(NULL != fileInfo);

    struct recls_strptrs_t file =
    {
            fileInfo->fileName.begin /* File is defined by start of fileName ... */
        ,   fileInfo->fileExt.end    /* ... to end of fileExt. */
    };

    return Recls_GetStringProperty_(&file, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetFileNameProperty( recls_info_t    fileInfo
                                            ,   recls_char_t    *buffer
                                            ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetFileNameProperty");

    recls_assert(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->fileName, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetFileExtProperty(  recls_info_t    fileInfo
                                            ,   recls_char_t    *buffer
                                            ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetFileExtProperty");

    recls_assert(NULL != fileInfo);

    return Recls_GetStringProperty_(&fileInfo->fileExt, buffer, cchBuffer);
}

RECLS_FNDECL(size_t) Recls_GetDirectoryPartProperty(    recls_info_t    fileInfo
                                                    ,   int             part
                                                    ,   recls_char_t    *buffer
                                                    ,   size_t          cchBuffer)
{
    function_scope_trace("Recls_GetDirectoryPartProperty");

    recls_assert(NULL != fileInfo);

    size_t  cParts = fileInfo->directoryParts.end - fileInfo->directoryParts.begin;

    if(part < 0)
    {
        return cParts;
    }
    else
    {
        recls_assert(static_cast<size_t>(part) < cParts);

        return Recls_GetStringProperty_(&fileInfo->directoryParts.begin[part], buffer, cchBuffer);
    }
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileReadOnly(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileReadOnly");

    recls_assert(NULL != fileInfo);

#if defined(RECLS_PLATFORM_IS_UNIX)
    return (fileInfo->attributes & S_IWRITE) == 0;
#elif defined(RECLS_PLATFORM_IS_WIN32)
    return fileInfo->attributes & FILE_ATTRIBUTE_READONLY;
#else /* unrecognised platform */
# error platform is not recognised
#endif /* platform */
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileDirectory(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileDirectory");

    recls_assert(NULL != fileInfo);

#if defined(RECLS_PLATFORM_IS_UNIX)
    return (fileInfo->attributes & S_IFMT) == S_IFDIR;
#elif defined(RECLS_PLATFORM_IS_WIN32)
    return fileInfo->attributes & FILE_ATTRIBUTE_DIRECTORY;
#else /* unrecognised platform */
# error platform is not recognised
#endif /* platform */
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileLink(recls_info_t fileInfo)
{
    function_scope_trace("Recls_IsFileLink");

    recls_assert(NULL != fileInfo);

#if defined(RECLS_PLATFORM_IS_UNIX) && \
    (   !defined(WIN32) || \
        !defined(EMULATE_UNIX_ON_WIN32))
    return (fileInfo->attributes & S_IFMT) == S_IFLNK;
#else /* unrecognised platform */
    STLSOFT_SUPPRESS_UNUSED(fileInfo);

    return false;
#endif /* platform */
}

RECLS_FNDECL(recls_bool_t) Recls_IsFileUNC(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetCreationTime");

    recls_assert(NULL != fileInfo);

#if defined(RECLS_PLATFORM_IS_WIN32) || \
    (   defined(RECLS_PLATFORM_IS_UNIX) && \
        defined(WIN32) && \
        defined(EMULATE_UNIX_ON_WIN32))
    return ('\\' == fileInfo->path.begin[0] && '\\' == fileInfo->path.begin[1]);
#else /* unrecognised platform */
    STLSOFT_SUPPRESS_UNUSED(fileInfo);

    return false;
#endif /* platform */
}

RECLS_FNDECL(recls_time_t) Recls_GetCreationTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetCreationTime");

    recls_assert(NULL != fileInfo);

    return fileInfo->_GetCreationTime;
}

RECLS_FNDECL(recls_time_t) Recls_GetLastStatusChangeTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetLastStatusChangeTime");

    recls_assert(NULL != fileInfo);

    return fileInfo->_GetLastStatusChangeTime;
}

RECLS_FNDECL(void) Recls_GetSizeProperty(   recls_info_t        fileInfo
                                        ,   recls_filesize_t    *size)
{
    function_scope_trace("Recls_GetSizeProperty");

    recls_assert(NULL != fileInfo);
    recls_assert(NULL != size);

    *size = fileInfo->size;
}

RECLS_FNDECL(recls_time_t) Recls_GetModificationTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetModificationTime");

    recls_assert(NULL != fileInfo);

    return fileInfo->modificationTime;
}

RECLS_FNDECL(recls_time_t) Recls_GetLastAccessTime(recls_info_t fileInfo)
{
    function_scope_trace("Recls_GetLastAccessTime");

    recls_assert(NULL != fileInfo);

    return fileInfo->lastAccessTime;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Roots
 */

RECLS_LINKAGE_C recls_char_t const *Recls_GetPathNameSeparator(void)
{
    static recls_char_t const   s_pathNameSeparator[] = { RAPI_PATHNAMESEP, '\0' };

    return s_pathNameSeparator;
}

RECLS_LINKAGE_C recls_char_t const *Recls_GetPathSeparator(void)
{
    static recls_char_t const   s_pathSeparator[] = { RAPI_PATHSEP, '\0' };

    return s_pathSeparator;
}

RECLS_LINKAGE_C recls_char_t const *Recls_GetWildcardsAll(void)
{
    static recls_char_t const   s_wildcardsAll[] = { RAPI_WILDCARDSALL, '\0' };

    return s_wildcardsAll;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Extended API functions
 */

static int RECLS_CALLCONV_DEFAULT IsDirectoryEmpty_proc(recls_info_t /* info */, recls_process_fn_param_t /* param */)
{
    return 0; // Cancel on any entry
}

RECLS_FNDECL(recls_bool_t) Recls_IsDirectoryEmpty(recls_char_t const *dir)
{
    recls_rc_t  rc  =   Recls_SearchProcess(dir, Recls_GetWildcardsAll(), RECLS_F_TYPEMASK, IsDirectoryEmpty_proc, 0);

    return RECLS_RC_SEARCH_CANCELLED != rc;
}

RECLS_FNDECL(recls_bool_t) Recls_IsDirectoryEntryEmpty(recls_info_t hEntry)
{
    recls_assert(NULL != hEntry);
    recls_assert(Recls_IsFileDirectory(hEntry));

    return Recls_IsDirectoryEmpty(hEntry->path.begin);
}

static int RECLS_CALLCONV_DEFAULT CalcDirectorySize_proc(recls_info_t hEntry, recls_process_fn_param_t param)
{
    recls_filesize_t    &total = *static_cast<recls_filesize_t*>(param);
    recls_filesize_t    size;

    Recls_GetSizeProperty(hEntry, &size);

#if defined(RECLS_PLATFORM_IS_UNIX)
    total += size;
#elif defined(RECLS_PLATFORM_IS_WIN32)
    total.QuadPart += size.QuadPart;
#else /* ? platform */
# error Platform not (yet) discriminated
#endif /* platform */

    return 1; // Never cancel
}

RECLS_FNDECL(recls_filesize_t) Recls_CalcDirectorySize(recls_char_t const *dir)
{
    static recls_filesize_t zero;
    recls_filesize_t        total = zero;

    Recls_SearchProcess(dir, Recls_GetWildcardsAll(), RECLS_F_FILES | RECLS_F_RECURSIVE, CalcDirectorySize_proc, static_cast<recls_process_fn_param_t>(&total));

    return total;
}

RECLS_FNDECL(recls_filesize_t) Recls_CalcDirectoryEntrySize(recls_info_t hEntry)
{
    recls_assert(NULL != hEntry);
    recls_assert(Recls_IsFileDirectory(hEntry));

    return Recls_CalcDirectorySize(hEntry->path.begin);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
