/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_util.cpp
 *
 * Purpose:     Platform-independent utility functions for the recls API.
 *
 * Created:     17th August 2003
 * Updated:     20th May 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls.h>
#include <recls_assert.h>
#include "recls_impl.h"
#include "recls_util.h"

#include <stlsoft.h>
#if !defined(_STLSOFT_VER_1_8_3)
# error STLSoft version 1.8.3 or greater required. (www.stlsoft.org/downloads.html)
#endif /* STLSoft version */

#if !defined(RECLS_COMPILER_IS_VECTORC)
# include <stlsoft_nulldef.h>
#endif /* compiler */

#if 0
# undef _RECLS_SUPPORTS_MULTIPATTERN
#endif /* 1 */


//#define RECLS_MULTIPATTERN_SEARCH_MANUAL

#ifndef RECLS_MULTIPATTERN_SEARCH_MANUAL

# ifndef RECLS_NO_USE_STRING_VIEW_FOR_PATTERN_SEARCH
#  define USE_STRING_VIEW_FOR_PATTERN_SEARCH
# endif /* !RECLS_NO_USE_STRING_VIEW_FOR_PATTERN_SEARCH */

# ifdef _RECLS_SUPPORTS_MULTIPATTERN
#  ifdef USE_STRING_VIEW_FOR_PATTERN_SEARCH
#   include <stlsoft/string_view.hpp>
#  else /* ? USE_STRING_VIEW_FOR_PATTERN_SEARCH */
#   include "recls_string.h"
#  endif /* USE_STRING_VIEW_FOR_PATTERN_SEARCH */
#  include <stlsoft/string_tokeniser.hpp>
# endif /* _RECLS_SUPPORTS_MULTIPATTERN */

#include <algorithm>    // for std::find

#endif /* !RECLS_MULTIPATTERN_SEARCH_MANUAL */

#include "recls_debug.h"

/* /////////////////////////////////////////////////////////////////////////////
 * Compiler compatibility
 */

#if defined(RECLS_COMPILER_IS_MSVC)
# if _MSC_VER < 1300
#  pragma warning(disable : 4127)
# endif /* _MSC_VER < 1300 */
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */

static const recls_char_t   s_localDir[]    =   RECLS_LITERAL(".");
static const recls_char_t   s_parentDir[]   =   RECLS_LITERAL("..");

/* ////////////////////////////////////////////////////////////////////////// */

#if defined(RECLS_CHAR_TYPE_IS_CHAR)
# define recls_strncpy__    strncpy
# define recls_strlen__     strlen
#elif defined(RECLS_CHAR_TYPE_IS_WCHAR)
# define recls_strncpy__    wcsncpy
# define recls_strlen__     wcslen
#else /* ? RECLS_CHAR_TYPE_IS_???? */
# error Unrecognised char type
#endif /* RECLS_CHAR_TYPE_IS_???? */

/* ////////////////////////////////////////////////////////////////////////// */

size_t recls_strncpy(recls_char_t *dest, size_t cchDest, recls_char_t const *src, size_t cchSrc)
{
    size_t  cchWritten;

    if(cchDest < cchSrc)
    {
        /* Just to straight strncpy. */
        recls_strncpy__(dest, src, cchDest);

        cchWritten = cchDest;
    }
    else
    {
        recls_strncpy__(dest, src, cchSrc);

        if(cchSrc < cchDest)
        {
            /* Fill the rest up with blanks. */
            memset(&dest[cchSrc], 0, sizeof(recls_char_t) * (cchDest - cchSrc));
        }

        cchWritten = cchSrc;
    }

    return cchWritten;
}

size_t recls_strlen(recls_char_t const *s)
{
    return recls_strlen__(s);
}

RECLS_LINKAGE_C size_t align_up_size(size_t i)
{
#ifdef RECLS_COMPILER_IS_BORLAND
# define cbAlign            sizeof(void*)
#else /* ? compiler */
    const size_t cbAlign =  sizeof(void*);
#endif /* RECLS_COMPILER_IS_BORLAND */

    return static_cast<size_t>((i + (cbAlign - 1)) & ~(cbAlign - 1));
}

RECLS_LINKAGE_C size_t count_char_instances(recls_char_t const *begin, recls_char_t const *end, recls_char_t const ch)
{
    recls_assert(NULL != begin);
    recls_assert(NULL != end);

    size_t  cDirParts   =   0;

    for(; begin != end; ++begin)
    {
        if(*begin == ch)
        {
            ++cDirParts;
        }
    }

    return cDirParts;
}

/* ////////////////////////////////////////////////////////////////////////// */

#if defined(RECLS_COMPILER_IS_DMC) || \
    defined(RECLS_COMPILER_IS_WATCOM)
RECLS_FNDECL(size_t) Recls_GetStringProperty_(  struct recls_strptrs_t const *      ptrs
                                            ,   recls_char_t *                      buffer
                                            ,   size_t                              cchBuffer)
#else
RECLS_FNDECL(size_t) Recls_GetStringProperty_(  struct recls_strptrs_t const *const ptrs
                                            ,   recls_char_t *const                 buffer
                                            ,   size_t const                        cchBuffer)
#endif /* RECLS_COMPILER_IS_DMC || RECLS_COMPILER_IS_WATCOM */
{
    recls_assert(NULL != ptrs);

    size_t  cch =   ptrs->end - ptrs->begin;

    if(NULL != buffer)
    {
        cch = recls_strncpy(buffer, cchBuffer, ptrs->begin, cch);
    }

    return cch;
}

/* ////////////////////////////////////////////////////////////////////////// */

RECLS_FNDECL(recls_rc_t) IsValidPattern(recls_char_t const *pattern, recls_uint32_t flags)
{
    recls_assert(NULL != pattern);

    if('\0' == *pattern)
    {
        return RECLS_RC_NO_MORE_DATA;
    }
    else
    {
#if !defined(_RECLS_SUPPORTS_MULTIPATTERN)

        STLSOFT_SUPPRESS_UNUSED(flags);
        STLSOFT_SUPPRESS_UNUSED(s_localDir);
        STLSOFT_SUPPRESS_UNUSED(s_parentDir);

#else /* ? _RECLS_SUPPORTS_MULTIPATTERN */
        // Ensure that if '.' was specified as a/the pattern, it is not
        // for recursive searches.

# ifdef RECLS_MULTIPATTERN_SEARCH_MANUAL

        STLSOFT_SUPPRESS_UNUSED(flags);
        STLSOFT_SUPPRESS_UNUSED(s_localDir);
        STLSOFT_SUPPRESS_UNUSED(s_parentDir);

        const recls_char_t          sep =   *Recls_GetPathSeparator();
        const size_t                len =   recls_strlen(pattern);
        recls_char_t const  *const  end =   pattern + len;

        if(1 == len)
        {
            if( 0 != (flags & RECLS_F_RECURSIVE) &&
                '.' == pattern[0])
            {
                return RECLS_RC_DOT_RECURSIVE_SEARCH;
            }
        }
        else if(2 == len)
        {
            if( '.' == pattern[0] && 
                '.' == pattern[1])
            {
                return RECLS_RC_DOT_RECURSIVE_SEARCH;
            }
        }

        if(len > 0)
        {
            recls_char_t    dotPattern[4]       =   "?.?";
            recls_char_t    dotdotPattern[5]    =   "?..?";

            dotPattern[0] = dotPattern[2] = dotdotPattern[0] = dotdotPattern[3] = sep;

            if( (   0 != (flags & RECLS_F_RECURSIVE) &&
                    0 == strncmp(pattern, &dotPattern[1], 2)) ||
                0 == strncmp(pattern, &dotdotPattern[1], 3))
            {
                return RECLS_RC_DOT_RECURSIVE_SEARCH;
            }

            if( (   0 != (flags & RECLS_F_RECURSIVE) &&
                    0 == strncmp(end - 2, dotPattern, 2)) ||
                0 == strncmp(end - 3, dotdotPattern, 3))
            {
                return RECLS_RC_DOT_RECURSIVE_SEARCH;
            }

            if( (   0 != (flags & RECLS_F_RECURSIVE) &&
                    NULL != strstr(pattern, dotPattern)) ||
                NULL != strstr(pattern, dotdotPattern))
            {
                return RECLS_RC_DOT_RECURSIVE_SEARCH;
            }
        }
# else  /* ? RECLS_MULTIPATTERN_SEARCH_MANUAL */

        // "." can only be specified as a pattern in non-recursive searches
# ifdef USE_STRING_VIEW_FOR_PATTERN_SEARCH
        typedef ::stlsoft::basic_string_view<recls_char_t>  string_t;
# else /* ? USE_STRING_VIEW_FOR_PATTERN_SEARCH */
        typedef RECLS_STRING_TEMPLATE_1(recls_char_t)       string_t;
# endif /* USE_STRING_VIEW_FOR_PATTERN_SEARCH */

        // Some compilers can't handle the instantiation of STLSoft's string_tokeniser
        // with a string delimiter type (as opposed to character), so for those we
        // stick with the character delimiter type
# if defined(RECLS_COMPILER_IS_WATCOM) || \
    (   defined(RECLS_COMPILER_IS_MSVC) && \
        _MSC_VER < 1300)
        typedef ::stlsoft::string_tokeniser<string_t, char>         tokeniser_t;

        tokeniser_t     tokens(pattern, *Recls_GetPathSeparator());
# else /* ? compiler */
        typedef ::stlsoft::string_tokeniser<string_t, string_t>     tokeniser_t;

        tokeniser_t     tokens(pattern, Recls_GetPathSeparator());
# endif /* compiler */

        if( tokens.end() != stlsoft_ns_qual_std(find)(tokens.begin(), tokens.end(), string_t(s_parentDir)) ||
            (   0 != (flags & RECLS_F_RECURSIVE) &&
                tokens.end() != stlsoft_ns_qual_std(find)(tokens.begin(), tokens.end(), string_t(s_localDir))))
        {
            return RECLS_RC_DOT_RECURSIVE_SEARCH;
        }

# endif  /* ? RECLS_MULTIPATTERN_SEARCH_MANUAL */
#endif /* _RECLS_SUPPORTS_MULTIPATTERN */

        return RECLS_RC_OK;
    }
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
